--- @brief Sends a character to a random position in the current scene.
---
--- @details
--- This will generate a random position in the scene, and check if the position is within a
--- certain area around the character's current position, and if it's inside the scene's current
--- current way system's border. If those conditions are met, the character is sent to this
--- new position.
---
--- @param char The character to move
--- @param maxDistance A `t_point` value with the maximum allowed X and Y distance to the character's current position (to prevent the character from walking too far away)
function sendCharacterToRandomPosition(char, maxDistance)
    if char == emptyObject or char == nil then
        char = game.CurrentCharacter
    end

    local currentScene = game.CurrentScene
    local charPosition = char.Position
    local sceneSize = currentScene.Sprite:getSize()
    local currentWaySystemPoints = currentScene.CurrentWaySystem.Border

    local validPositionFound = false
    local position = { x = 0, y = 0}
    while not validPositionFound do
        -- New random position
        local randomPosition = { x = math.random(0, sceneSize.x), y = math.random(0, sceneSize.y) }

        -- Check distance of random position to char position
        if (math.abs(charPosition.x - randomPosition.x) <= maxDistance.x) and (math.abs(charPosition.y - randomPosition.y) <= maxDistance.y) then
            -- Check if random position is inside the walkable area
            if isPointInsidePolygon(randomPosition, currentWaySystemPoints) then
                position = randomPosition
                validPositionFound = true
            end
        end
    end

    -- Send character to new position
    --print("Sending character " .. char:getName() .. " to position " .. vect_to_string(position) .. "...")
    char.Destination = position

    -- If "wait" is true
    if doWait == 1 then
        game.HideCursor = true
        if not (char.Position.x == position.x and char.Position.y == position.y) then
            return true
        end
    end
    game.HideCursor = false
    return false
end

--- @brief Sends a character near a defined position in the current scene.
---
--- @details
--- This will move the character `char` towards `targetPos`, but keep a distance of `distance` to it.
---
--- @param char The character to move
--- @param targetPos A `t_point` value with the target position to move near to
--- @param distance The distance towards targetPos to keep
function sendCharacterToNearPosition(char, targetPos, distance, doWait)
    if char == emptyObject or char == nil then
        char = game.CurrentCharacter
    end

    local currentScene = game.CurrentScene
    local charPosition = char.Position
    local currentWaySystemPoints = currentScene.CurrentWaySystem.Border

    -- Compute new position
    local newPosition = vect_add(targetPos, vect_mul_scalar(vect_normalize(vect_subtract(charPosition, targetPos)), distance))

    -- Check if new position is inside the walkable area
    if isPointInsidePolygon(newPosition, currentWaySystemPoints) then
        -- Send character to new position
        --print("Sending character " .. char:getName() .. " to position " .. vect_to_string(newPosition) .. "...")
        char.Destination = newPosition

        -- If "wait" is true
        if doWait == 1 then
            game.HideCursor = true
            if not (char.Position.x == newPosition.x and char.Position.y == newPosition.y) then
                return true
            end
        end
    end
    game.HideCursor = false
    return false
end

--- @brief Sends a character near a specified object or character in the current scene.
---
--- @details
--- This will move the character `char` towards `targetPos`, but keep a distance of `distance` to it.
---
--- @param char The character to move
--- @param targetObj An object or character to move near to
--- @param distance The distance towards targetObj to keep
--- @param offset Offset the computed destination position by this vector
--- @param doWait If `1`, the action part will wait until the character has reached their destination position.
--- @param useDestinationPosition If `0`, `targetObj.Position` will be used. If `1`, `targetObject.Destination` will be used.
function sendCharacterToNearObjectPos(char, targetObj, distance, offset, doWait, useDestinationPosition)
    if char == emptyObject or char == nil then
        char = game.CurrentCharacter
    end

    if targetObj ~= emptyObject and targetObj ~= nil and char ~= targetObj then
        useDestinationPosition = useDestinationPosition or 0
        local currentScene = game.CurrentScene
        local charPosition = char.Position
        local targetPosition = targetObj.Position
        if useDestinationPosition then
            targetPosition = targetObj.Destination
        end
        local currentWaySystemPoints = currentScene.CurrentWaySystem.Border

        --print("--")
        --print("Character position: " .. vect_to_string(charPosition))
        --print("Target object position: " .. vect_to_string(targetPosition))
        --print("Offset: " .. vect_to_string(offset))

        -- Compute new position
        local newPosition = char.Position
        -- Only apply distance if character is far enough away from the target
        if vect_distance(newPosition, targetPosition) > distance then
            newPosition = vect_add(targetPosition, vect_mul_scalar(vect_normalize(vect_subtract(charPosition, targetPosition)), distance))
        end
        --print("Destination position: " .. vect_to_string(newPosition))

        -- Check if new position is inside the walkable area
        if isPointInsidePolygon(newPosition, currentWaySystemPoints) then
            -- Add offset
            newPosition.x = newPosition.x + offset.x
            newPosition.y = newPosition.y + offset.y
            --print("Destination position with offset: " .. vect_to_string(newPosition))
            
            -- Assign new destination position
            --print("Sending character " .. char:getName() .. " to position " .. vect_to_string(newPosition) .. "...")
            char.Destination = newPosition

            -- If "wait" is true
            if doWait == 1 then
                game.HideCursor = true
                --if not (char.Position.x == newPosition.x and char.Position.y == newPosition.y) then -- Esmeralda's way of waiting until destination is reached.
                if char.AnimState == 1 then -- I find this method better, as it avoids deadlocks in case the character can't reach its destination for some reason.
                    return true
                end
            end
        end
    end
    game.HideCursor = false
    return false
end


-- Register action parts
registerActionPart('SENDCHARTORNDPOS', function(activeAction, actionPart)
    return sendCharacterToRandomPosition(actionPart.Link, { x = actionPart.Int, y = actionPart.AltInt}, false)
end)
registerActionPart('SENDCHARNEARPOS', function(activeAction, actionPart)
    return sendCharacterToNearPosition(actionPart.Link, { x = actionPart.Int, y = actionPart.AltInt}, actionPart.AltInt2)
end)
registerActionPart('SENDCHARNEAROBJ', function(activeAction, actionPart)
    return sendCharacterToNearObjectPos(actionPart.Link, actionPart.AltLink, actionPart.Int, { x = 0, y = 0}, actionPart.AltInt2)
end)
registerActionPart('SENDCHARNEARCHAR', function(activeAction, actionPart)
    return sendCharacterToNearObjectPos(actionPart.Link, actionPart.AltLink, actionPart.Int, { x = 0, y = 0}, actionPart.AltInt2, 1)
end)
registerActionPart('SENDCHARRELTOCHAR', function(activeAction, actionPart)
    return sendCharacterToNearObjectPos(actionPart.Link, actionPart.AltLink, 0, { x = actionPart.Int, y = actionPart.AltInt}, actionPart.AltInt2, 1)
end)
registerActionPart('SENDCHARRELTOOBJ', function(activeAction, actionPart)
    return sendCharacterToNearObjectPos(actionPart.Link, actionPart.AltLink, 0, { x = actionPart.Int, y = actionPart.AltInt}, actionPart.AltInt2)
end)
